/*$Id: e_node.h,v 25.6 2004/04/28 12:12:59 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@ieee.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *------------------------------------------------------------------
 * circuit node class
 */
#ifndef E_NODE_H
#define E_NODE_H
#include "u_status.h" // time critical inline
#include "u_opt.h" // time critical inline
#include "constant.h"
#include "e_base.h"
/*--------------------------------------------------------------------------*/
class MODEL_LOGIC;
class XPROBE;
/*--------------------------------------------------------------------------*/
enum {
  OUT1 = 0,
  OUT2 = 1,
  IN1 = 2,
  IN2 = 3,
  NODES_PER_BRANCH = 4,
  INVALID_NODE = -1
};
#define	qBAD	 (0)
#define qGOOD	 (OPT::transits)
/*--------------------------------------------------------------------------*/
enum _LOGICVAL {lvSTABLE0,lvRISING,lvFALLING,lvSTABLE1,lvUNKNOWN};
enum {lvNUM_STATES = lvUNKNOWN+1};
/*--------------------------------------------------------------------------*/
class LOGICVAL {
private:
  _LOGICVAL _lv;
  static const _LOGICVAL or_truth[lvNUM_STATES][lvNUM_STATES];
  static const _LOGICVAL xor_truth[lvNUM_STATES][lvNUM_STATES];
  static const _LOGICVAL and_truth[lvNUM_STATES][lvNUM_STATES];
  static const _LOGICVAL not_truth[lvNUM_STATES];
public:
  LOGICVAL() :_lv(lvUNKNOWN)			{}
  LOGICVAL(const LOGICVAL& p)	:_lv(p._lv)	{}
  LOGICVAL(_LOGICVAL p)		:_lv(p)		{}

  operator _LOGICVAL()const {return static_cast<_LOGICVAL>(_lv);}
  
  LOGICVAL& operator=(_LOGICVAL p)	{_lv=p; return *this;}
  LOGICVAL& operator=(const LOGICVAL& p) {_lv=p._lv; return *this;}

  LOGICVAL& operator&=(LOGICVAL p)
    {untested(); _lv = and_truth[_lv][p._lv]; return *this;}
  LOGICVAL& operator|=(LOGICVAL p) {_lv = or_truth[_lv][p._lv]; return *this;}
  LOGICVAL  operator^=(LOGICVAL p)
    {untested(); _lv = xor_truth[_lv][p._lv]; return *this;}
  LOGICVAL  operator~()const	   {return not_truth[_lv];}
  
  bool is_unknown()const{return _lv == lvUNKNOWN;}
  bool lv_future()const	{assert(_lv!=lvUNKNOWN); return _lv & 1;}
  bool lv_old()const	{assert(_lv!=lvUNKNOWN); return _lv & 2;}

  bool is_rising() const{return _lv == lvRISING;}
  bool is_falling()const{return _lv == lvFALLING;}
  LOGICVAL& set_in_transition(LOGICVAL newval);
};
/*--------------------------------------------------------------------------*/
class nODE : public CKT_BASE {
  friend class node_t;
private:
  int	      _user_number;	/* external node number (redundant) */
  const MODEL_LOGIC *_family;	/* logic family */
  int 	      _d_iter;		/* iteration of last update - digital */
  int 	      _a_iter;		/* iteration of last update - analog */
  double      _final_time;	/* time logic transition attains final state */
  double      _lastchange;	/* time of last change */
  double      _old_lastchange;	/* in case it rejects a step */
  smode_t     _mode;		/* simulation mode (analog or digital)*/
  LOGICVAL    _lv;		/* "logic value" (real type is LOGICVAL) */
  LOGICVAL    _old_lv;		/* in case it rejects a step */
  int	      _quality;		/* quality of digital mode */
  const char* _failure_mode;

  // so it is not pure virtual
  const	      std::string long_label()const
			{unreachable(); return "unreachable";}
private:
  double      tr_probe_num(CS&)const;
  XPROBE      ac_probe_ext(CS&)const;
  double      annotated_logic_value()const;
  //nODE&	      set_bad_quality(const char* f)
  //			{_quality = qBAD; _failure_mode = f; return *this;}
  int	      user_number()const	{return _user_number;}
  int	      internal_number()const	{return SIM::nm[_user_number];}
  int	      quality()const		{return _quality;}
  double      final_time()const		{return _final_time;}
  double      v0()const			{return SIM::v0 [internal_number()];}
  COMPLEX     vac()const		{return SIM::ac [internal_number()];}

private: // inhibited
  explicit    nODE(const nODE& p)	:CKT_BASE(p) {unreachable();}
public: // general use
  explicit    nODE();
	      ~nODE()			{}

public: // init
  nODE&	      set_number(int n)		{_user_number = n; return *this;}
public: // matrix
  nODE&	      set_a_iter()	{_a_iter = STATUS::iter[iTOTAL]; return *this;}
};
/*--------------------------------------------------------------------------*/
class node_t {
private:
  static int* _analog_fanout;
public:
  static bool node_is_valid(int i)
		{return i>=0 && i<=STATUS::total_nodes;}
  static int  to_internal(int n)
		{assert(node_is_valid(n)); assert(SIM::nm); return SIM::nm[n];}
  static int  analog_fanout(int n)
  {assert(node_is_valid(n)); assert(_analog_fanout); return _analog_fanout[n];}
  static bool has_analog(int n)
		{return analog_fanout(n) > 0;}

private:
  int _m;		// mapped, after reordering
  int _t;		// m == nm[t] if properly set up
  int _e;		// external, user numbers, t = subs exp

public:
  int	      m_()const 		{return _m;}
  int	      t_()const 		{return _t;} // d_subckt.cc:271 only
  int	      e_()const			{return _e;}

  std::string name(const CARD*)const;
  void	      parse(CS&, const CARD*);
  void	      set_to_0()		{_t = _e = 0;}
  void	      new_model_node()		{_t = STATUS::newnode_model();}
  void	      map_subckt_node(int* map_array);
  bool	      is_valid()const
		{return node_is_valid(_t) && _m==to_internal(_t);}
  bool	      is_unconnected()const	{return (_e == INVALID_NODE);}
  node_t&     map()			{_m=to_internal(_t); return *this;}

  explicit    node_t() : _m(INVALID_NODE),_t(INVALID_NODE),_e(INVALID_NODE) {}
	      node_t(const node_t& p)	:_m(p._m), _t(p._t), _e(p._e) {}
  explicit    node_t(int ee)	    :_m(INVALID_NODE), _t(ee), _e(ee) {map();}

  static void allocate(int n);
  static void unallocate();

  bool	      operator==(const node_t& p)const  {return _t == p._t;}
  bool	      operator!=(const node_t& p)const  {return !(*this == p);}
  node_t&     operator=(int n)		{_t=_e=n; return map();}

private: // raw data access (lvalues)
  nODE&	      data()const
    {assert(is_valid());extern nODE*nstat;assert(nstat);return nstat[m_()];}

private: // raw data access (lvalues)
  void	set_quality(int q)		{data()._quality = q;}
  void	set_failure_mode(const char* f) {data()._failure_mode = f;}
  void	set_final_time(double t)	{data()._final_time = t;}
  
  void	set_d_iter()		      {data()._d_iter = STATUS::iter[iTOTAL];}
  void	set_last_change_time()		{data()._lastchange = SIM::time0;}
  void	set_last_change_time(double t)	{data()._lastchange = t;}
  void	set_lv(LOGICVAL v)		{data()._lv = v;}
  void	set_process(const MODEL_LOGIC* f) {data()._family = f;}

  void store_old_last_change_time(){data()._old_lastchange=last_change_time();}
  void	store_old_lv()			{data()._old_lv = lv();}
  void	restore_lv()			{untested(); set_lv(old_lv());}

private: // raw data access (rvalues)
  bool	is_digital()const		{return data()._mode == moDIGITAL;}
  bool	is_analog()const		{return data()._mode == moANALOG;}
  const MODEL_LOGIC* process()const	{return data()._family;}
  double old_last_change_time()const {untested(); return data()._old_lastchange;}
  LOGICVAL old_lv()const		{return data()._old_lv;}

public: // raw data access (rvalues)
  int	      quality()const		{return data()._quality;}
  const char* failure_mode()const	{return data()._failure_mode;}
  int	      d_iter()const		{return data()._d_iter;}
  int	      a_iter()const		{return data()._a_iter;}
  double      final_time()const		{return data()._final_time;}
  double      last_change_time()const	{return data()._lastchange;}
  bool        lv_future()const		{return lv().lv_future();}

private: // other internal
  void  set_bad_quality(const char* f) {set_quality(qBAD);set_failure_mode(f);}
  void  set_good_quality(const char* f = "ok")
				{set_quality(qGOOD); set_failure_mode(f);}
  void	dont_set_quality(const char* f = "don't know") {set_failure_mode(f);}
  void	improve_quality()	{if (quality() < qGOOD) {++data()._quality;}}
  bool	just_reached_stable()const;

private: // internal data access
  const	      std::string long_label()const;

public: // data access, used by logic
  bool	      in_transit()const		{return final_time() < NEVER;}

public: // action, used by logic
  void	      set_event(double delay, LOGICVAL v);
  void	      force_initial_value(LOGICVAL v);
  void	      propagate();
  double      to_analog(const MODEL_LOGIC*f);
  void	      to_logic(const MODEL_LOGIC*f);
  void	      set_mode(smode_t m)	{data()._mode = m;}

public:
  LOGICVAL    lv()const			{return data()._lv;}
  bool	      is_unknown()const		{return lv().is_unknown();}
  double      tr_probe_num(CS& s)const;
  XPROBE      ac_probe_ext(CS& s)const;
  void	      inc_analog_fanout()	{
    assert(is_valid());
    assert(_analog_fanout);
    ++_analog_fanout[m_()];
  }

  double      v0()const			{return SIM::v0[m_()];}	/* rvalues */
  double      vt1()const		{return SIM::vt1[m_()];}
  COMPLEX     vac()const		{return SIM::ac[m_()];}
  double&     i()			{return SIM::i[m_()];}	/* lvalues */
  COMPLEX&    iac()			{return SIM::ac[m_()];}

#if 0
  double      vdc()const		{return SIM::vdc[m_()];}
#endif
  int analog_fanout()const {
    untested();
    assert(is_valid());
    assert(_analog_fanout);
    return _analog_fanout[m_()];
  }


  bool has_analog()const {
    untested();
    return analog_fanout() > 0;
  }
};
OMSTREAM& operator<<(OMSTREAM& o, const node_t& n);
/*--------------------------------------------------------------------------*/
double volts_limited(const node_t& n1, const node_t& n2);
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
#endif
